{% extends "!layout.html" %}

{% block rootrellink %}
        <li><a href="{{ pathto('index') }}">home</a>|&nbsp;</li>
        <li><a href="{{ pathto('search') }}">search</a>|&nbsp;</li>
	<li><a href="{{ pathto('contents') }}">contents</a> &raquo;</li>
{% endblock %}




